/**
 * Created by Ajay Kulkarni 6.7.2001
 *
 *
 */

package agentland.behavior;

import metaglue.*;
import java.rmi.*;
import agentland.resource.*;
import agentland.util.*;
import agentland.behavior.speechin.*;
import java.util.*;
import java.lang.reflect.*;


/**
 * <code>BehaviorAgent</code> is the parent class for all
 * Behaviors.
 *
 * Changes I will be making soon:<br>
 * -> Use persistent maps and other forms of persistent storage<br>
 *
 * Ajay Kulkarni 7.25.2001
 *
 * @see ManagedAgent
 * @see Behavior
 */
public class BehaviorAgent extends ManagedAgent implements Behavior {

    // behavior states
    public static final int INACTIVE = 0;
    public static final int LISTENING = 1;
    public static final int LEAF_ACTIVE = 2;
    public static final int OVERRIDEN_ACTIVE = 3;
    
    // Internal Representation 
    private boolean DEBUG = true;
    
    // * Unique identifier for each Behavior
    // * Is set by Behavior Coordinator upon Behavior registration with 
    // Coordinator
    private BehaviorID ID;

    // Potential parents
    private Vector parentBevs = new Vector();

    // Current active parents AND ANCESTORS
    private Vector activeParents = new Vector();

    // Current Active Child-Parent Relationships
    // (i.e., active parents, NO ANCESTORS)
    private Vector activeDirectParents = new Vector();    

    // Current active child bevs
    // i.e., behaviors that will pass a secret directly to this bev
    private Vector childrenBevs = new Vector();


    // how many children have passed secrets to this bev
    // (this bev won't handle a secret until all children have
    // handled it already)
    // this var is reset after each secret is passed
    private int childPasses = 1;

    // Possible Actions (Vector of Method objects)
    private Vector actions = new Vector();

    // Overriden Actions
    private Vector inactiveActions = new Vector();

    // Temporarily Unoverriden Actions
    private Vector tempActiveActions = new Vector();

    // Actions that CANNOT be overriden
    private Vector permanentActions = new Vector();

    // Events it listens for
    //    private Vector events = new Vector();

    // Hashtable mapping the Events this Behavior listens for
    // to the Actions that act upon those events
    // (Event) => [(Action1), (Action2), ...]
    private Hashtable eventToActionsVectorHash = new Hashtable();

    // Events that the activate method listens for
    // note that this needs to be maintained separately from
    // the eventToActionsVectorHash, as activate is not an Action
    private Vector activationEvents = new Vector();
    
    // Current Active Parent  (DEPRECATED)
    private Behavior activeParent;

    // Current State of the Behavior
    //   0 = inactive
    //   1 = listening
    //   2 = active leaf
    //   3 = overriden active (i.e., non-leaf active)
    private int state = INACTIVE; 

    // This is the heart of the action-level dependency:
    // * Hashtable that, based on the behavioral ID of a parent, 
    // retrieves another hashtable
    // * This second hashtable hashes a Vector containing the names 
    // of parent actions with
    // the name of an action in this behavior
    // * "this action" overrides "parent action"
    // (I.E.: 
    //       * parentToOverridingActionsHash(parentID --> actionsHash)
    //       * actionsHash(thisAction --> parentActionsVector)
    //       * parentActionsVector: (parentAction1, parentAction2, etc.)
    private Hashtable parentToOverridingActionsHash = new Hashtable();

    // Behavior Coordinator that this behavior is connected to
    private BehaviorCoordinator bevCoord;

    // Look-up Table to translate agentIDs to agent stubs
    // this way, the number of reliesOn() calls are minimized
    private Hashtable agentIDToAgentStubHash = new Hashtable();

    // where the name of the action currently being invoked
    // is stored. this helps to allow an action to call the action 
    // that it is supposed to override
    // (is empty if no action is currently being invoked)
    private String currentAction = new String();

    // to allow users to give the same behavior various aliases
    private Vector aliases;

    // SPEECH STUFF:
    // internal rep for the behavior's speech agent
    protected BehaviorSpeech bevspeech;
    

    
    
    // Methods 
    

    /**
     * Creates a new <code>BehaviorAgent</code> instance.<br>
     * <br>
     * Contains the code that stores all the Actions of a particular
     * Behavior subclass. Must be called in each Behavior
     * subclass. (i.e., by calling <code>super();</code> in the
     * subclass constructor. 
     *
     * @exception RemoteException if an error occurs
     */
    
    public BehaviorAgent() throws RemoteException {
	super();
	setLogLevel(LogStream.DEBUG);

	// add actions
	Method[] declMethods = this.getClass().getDeclaredMethods();
	for (int i = 0; i < declMethods.length; i++) 
	    if (declMethods[i].getName().lastIndexOf("action") != -1)	
		addAction(declMethods[i]);
	ID = new BehaviorID(getAgentID());
	
	// unfreeze aliases
	Object o = defrost("aliases");
	if (o != null) 
	    aliases = (Vector) o;
	else
	    aliases = new Vector();

	// get behavior speech agent
	bevspeech = (BehaviorSpeech) reliesOn("agentland.behavior.speechin.BehaviorSpeech");

	log("DEBUG","Added Actions");
    }

    
    /**
     * Returns the BehaviorID of this Behavior.
     *
     * @return a <code>BehaviorID</code> value
     */
    
    public BehaviorID getID()  {
	return ID;
    }


    /**
     * How to set the numeric value of this Behavior's BehaviorID
     *
     * @see BehaviorID
     *
     * @param num an <code>Integer</code> value
     * @exception RemoteException if an error occurs
     */
    
    public void setIDNumeric(Integer num) throws RemoteException {
	ID.setNumeric(num);
    }


    /**
     * A Behavior can be in one of four states:<br>
     * 0: Inactive<br>
     * 1: Listening<br>
     * 2: Leaf Active<br>
     * 3: Overriden Active<br>
     *
     * Returns true if this Behavior is Overriden.<br>
     *
     * @return a <code>boolean</code> value
     */
    
    public boolean isOverriden() {
	return (state == OVERRIDEN_ACTIVE);
    }


    /**
     * A Behavior can be in one of four states:<br>
     * 0: Inactive<br>
     * 1: Listening<br>
     * 2: Leaf Active<br>
     * 3: Overriden Active<br>
     *
     * Returns true if this Behavior is a Leaf.<br>
     *
     * @return a <code>boolean</code> value
     */
    public boolean isLeaf() {
	return (state == LEAF_ACTIVE);
    }


    /**
     * A Behavior can be in one of four states:<br>
     * 0: Inactive<br>
     * 1: Listening<br>
     * 2: Leaf Active<br>
     * 3: Overriden Active<br>
     *
     * Returns true of this Behavior is Active.<br>
     *
     * @return a <code>boolean</code> value
     */
    public boolean isActive() {
	return (state >= LEAF_ACTIVE);
    }


    /**
     * A Behavior can be in one of four states:<br>
     * 0: Inactive<br>
     * 1: Listening<br>
     * 2: Leaf Active<br>
     * 3: Overriden Active<br>
     *
     * Returns true if this Behavior is Listening.<br>
     *
     * @return a <code>boolean</code> value
     */
    public boolean isListening() {
	return (state == LISTENING);
    }


    /**
     * A Behavior can be in one of four states:<br>
     * 0: Inactive<br>
     * 1: Listening<br>
     * 2: Leaf Active<br>
     * 3: Overriden Active<br>
     *
     * Returns true if this Behavior is Inactive.<br>
     *
     * @return a <code>boolean</code> value
     */
    public boolean isInactive()  {
	return (state == INACTIVE);
    }

    // NOTE: 
    // * these following "make" methods use the same
    // state <==> integer mapping as the bev coord
    // * if this were not the case, then you would need to do
    // more comparisons (in the following "make" methods)

    
    /**
     * Changes the state of the Behavior to Overriden.<br>
     * 
     * If this bev is registered with a Behavior Coordinator (as it
     * should be), it updates its state with the Coordinator.<br>
     *
     * @see BehaviorCoordinator
     *
     */
    protected void makeOverriden() {
	int oldstate = state; // state that bev is being changed from
	state = OVERRIDEN_ACTIVE; // setting state to overriden
	int newstate = state; // state that bev is being changed to
	try {
	    if (bevCoord != null) 
		bevCoord.updateBev(getID(), oldstate, newstate);
	} catch(RemoteException e) {}
    }

    
    /**
     * Changes the state of the Behavior to UnOverriden.(i.e., from
     * Overriden Active to Leaf Active)<br>
     * 
     * If this bev is registered with a Behavior Coordinator (as it
     * should be), it updates its state with the Coordinator.<br>
     *
     * @see BehaviorCoordinator
     *
     */
    protected void makeUnOverriden() {
	int oldstate = state; // state that bev is being changed from
	state = 2; // setting state to active leaf
	int newstate = state; // state that bev is being changed to
	try {
	    if (bevCoord != null) 
		bevCoord.updateBev(getID(), oldstate, newstate);
	} catch(RemoteException e) {}
    }
    

    /**
     * Changes the state of the Behavior to Leaf Active from a
     * non-active state.<br>
     * 
     * If this bev is registered with a Behavior Coordinator (as it
     * should be), it updates its state with the Coordinator.<br>
     *
     * This method also climbs up the active behavioral tree, identifying
     * the direct parent(s) up this behavior. 
     *
     * As the Behavior is transitioning from a non-active state to
     * leaf active, this method overrides the appropriate actions of
     * its active ancestors. 
     * 
     * @see BehaviorCoordinator
     *
     */
    protected void makeActive() {
	int oldstate = state; // state that bev is being changed from
	state = LEAF_ACTIVE; // setting state to active leaf
	int newstate = state; // state that bev is being changed to
	try {
	    if (bevCoord != null) 
		bevCoord.updateBev(getID(), oldstate, newstate);
	} catch(RemoteException e) {}

	try {
	    // refresh active parent list
	    // (you might have active parents that this
	    // bev doesn't know about,
	    // b/c they were activated after this bev 
	    // became listening)
	    Enumeration pars = getParents().elements();
	    while (pars.hasMoreElements()) {
		BehaviorID parID = (BehaviorID) pars.nextElement(); // get bevID
		Behavior par = (Behavior) agentIDToAgentStubHash.get(parID.getAgentID()); // look up agentID to get agent stub
		if (par == null) { 
		    // if agentID wasn't in hashtable, rely on agent and then 
		    // store in hashtable
		    par = (Behavior) reliesOn(parID.getAgentID());
		    agentIDToAgentStubHash.put(parID.getAgentID(), par);
		}
		    
		if (par.isActive()) {
		    if (getActiveParents().indexOf(parID) == -1) {
			activateParent(parID);
		    }
		}
	    }

	    
	    // Ajay's going to start commenting in Haikaii no Rengas...
	    // Haikaii No Renga: (5-7-5, 7-7, 5-7-5, 7-7, etc.)

	    // find direct parent:
	    // climb the direct parent tree
	    // start with the leaf bevs

	    // until you find a parent
	    // you must continue climbing

	    // (note that for most bevs,
	    // a leaf bev will be parent,
	    // and you will climb not)
	    if (bevCoord != null) {

		Vector nextLevel = new Vector();
		Enumeration leafBevs = bevCoord.getLeafBevs().elements();

		// while you don't find a direct active parent
		// climb the tree
		while (getDirectActiveParents().isEmpty()) {
		    
		    while (leafBevs.hasMoreElements()) {
			BehaviorID bevID = (BehaviorID) leafBevs.nextElement();

			// the bev coord has already made this
			// bev a leaf, but hasn't made the old
			// leaves into overriden bevs...
			// so you get both right now, but you
			// WANT the overriden bevs
			// so just get the leaf bevs that aren't this bev
			if (!bevID.equals(getID())) {
			    
			    if (isParent(bevID)) {
				// found a direct active parent
				activateDirectParent(bevID);
			    }
			    else { 
				Behavior bev = (Behavior) agentIDToAgentStubHash.get(bevID.getAgentID());
				if (bev == null) {
				    bev = (Behavior) reliesOn(bevID.getAgentID());
				    agentIDToAgentStubHash.put(bevID.getAgentID(), bev);
				}

				nextLevel.addAll(bev.getDirectActiveParents());
			    }
			}
		    }
		    
		    // next level of bevs to check out
		    leafBevs = nextLevel.elements();
		}
	    } else {
		log("ERROR", "Not registered with Bev Coord!");
	    }
	    

	    // Get a list of all the behavior IDs of all the behaviors
	    // that this behavior overrides

	    // note that these behaviors may not just be the "parents"
	    // of this bev: a bev may override actions of behaviors
	    // that are > 1 hierarchical level upwards
	    // (i.e., ancestors)
	    pars = parentToOverridingActionsHash.keys();
	    while (pars.hasMoreElements()) {
		BehaviorID parID = (BehaviorID) pars.nextElement(); // get bevID
		Behavior par = (Behavior) agentIDToAgentStubHash.get(parID.getAgentID()); // look up agentID to get agent stub
		if (par == null) { 
		    // if agentID wasn't in hashtable, rely on agent and then 
		    // store in hashtable
		    par = (Behavior) reliesOn(parID.getAgentID());
		    agentIDToAgentStubHash.put(parID.getAgentID(), par);
		}
		
		// For each behavior that's alive, send it the list
		// of action names that this behavior is overriding
		if (par.isActive()) {
	 	    Hashtable overs = (Hashtable) parentToOverridingActionsHash.get(parID);
		    Vector actionNames = new Vector();
		    Enumeration actionVectors = overs.elements();
		    while (actionVectors.hasMoreElements()) 
			actionNames.addAll((Vector)actionVectors.nextElement());
		    par.override(actionNames);
		}
	    }
	} catch (Exception e) {}
    }
    

    /**
     * Changes the state of the Behavior to Listening.<br>
     * 
     * If this bev is registered with a Behavior Coordinator (as it
     * should be), it updates its state with the Coordinator.<br>
     *
     * @see BehaviorCoordinator
     *
     */
    protected void makeListening() {
	int oldstate = state; // state that bev is being changed from
	state = LISTENING; // setting state to listening
	int newstate = state; // state that bev is being changed to
	try {
	    if (bevCoord != null) 
		bevCoord.updateBev(getID(), oldstate, newstate);
	} catch(RemoteException e) {}
    }


    /**
     * Changes the state of the Behavior to Inactive.<br>
     * 
     * If this bev is registered with a Behavior Coordinator (as it
     * should be), it updates its state with the Coordinator.<br>
     *
     * @see BehaviorCoordinator
     *
     */
    protected void makeInactive() {
	int oldstate = state; // state that bev is being changed from
	state = INACTIVE; // setting state to inactive
	int newstate = state; // state that bev is being changed to
	try {
	    if (bevCoord != null) 
		bevCoord.updateBev(getID(), oldstate, newstate);
	} catch(RemoteException e) {}
    }


    /**
     * This method is called when a Behavior is transitioning from an
     * Active state (leaf or overriden) to a Non-Active state
     * (listening or inactive). It contains tasks common for when a
     * Behavior is deactivated: e.g., undo the overriding of
     * ancestorial behaviors. <br>
     *
     * Note that while a Behavior used to have to explicitly call
     * this method, now, this method is called automatically by the 
     * pass() method if the behavior becomes deactivated from 
     * its actionDeactive method (i.e., its deactive action).
     *
     * @see #pass(Secret secret)
     *
     */
        protected void makeDeActive() {
	// common tasks when a behavior is deactivated
	
	try {
	    // Get a list of all the behavior IDs of all the behaviors
	    // that this behavior overrides
	    
	    // note that these behaviors may not just be the 
	    // "parents" of this bev: teh parent of a bev is 
	    // a bev that one hierarchical level upwards. 
	    // BUT a bev may override actions of behaviors that
	    // are > 1 hierarchical level upwards 
	    // (i.e., ancestors)
	    Enumeration pars = parentToOverridingActionsHash.keys();
	    while (pars.hasMoreElements()) {
		BehaviorID parID = (BehaviorID) pars.nextElement(); // get bevID
		Behavior par = (Behavior) agentIDToAgentStubHash.get(parID.getAgentID()); // look up agentID to get agent stub
		if (par == null) { 
		    // if agentID wasn't in hashtable, rely on agent and then 
		    // store in hashtable
		    par = (Behavior) reliesOn(parID.getAgentID());
		    agentIDToAgentStubHash.put(parID.getAgentID(), par);
		}

		// For each behavior that's alive, send it the list
		// of action names that this behavior used to override
		// and must now be reactivated
		if (par.isActive()) {

		    Hashtable overs = 
			(Hashtable) parentToOverridingActionsHash.get(parID);
		    Vector actionNames = new Vector();
		    Enumeration actionVectors = overs.elements();
		    while (actionVectors.hasMoreElements())
			actionNames.addAll((Vector)actionVectors.nextElement());
		    par.undoOverride(actionNames);
		}
	    }


	    // For each direct active parent behavior, tell it that you're
	    // being deactivated
	    pars = getDirectActiveParents().elements();
	    while (pars.hasMoreElements()) {
		Behavior par = (Behavior) agentIDToAgentStubHash.get(((BehaviorID)pars.nextElement()).getAgentID());
		par.removeChild(getID());
	    }

	    
	    // For each direct child behavior, tell it that you're
	    // being deactivated
	    Enumeration children = getChildren().elements();
	    while (children.hasMoreElements()) {
		Behavior child = (Behavior) agentIDToAgentStubHash.get(((BehaviorID)children.nextElement()).getAgentID());
		child.deactivateParent(getID());
	    }

	    // reset inactive actions list
	    inactiveActions = new Vector();

	    // reset direct active parents list
	    activeDirectParents = new Vector();

	    // reset children list
	    childrenBevs = new Vector();
	} catch (Exception e) {}
    }
	
	
    
    /**
     * Adds a child-parent relationship between this behavior (child)
     * and another behavior with ID <code>bID</code>.
     *
     * @param bID The <code>BehaviorID</code> of the parent
     * @exception RemoteException if an error occurs
     */
    public void addParent(BehaviorID bID) throws RemoteException {
	if (parentBevs.indexOf(bID) == -1) {
	    parentBevs.add(bID);

	    // store agent stub 
	    agentIDToAgentStubHash.put(bID.getAgentID(), (Behavior)reliesOn(bID.getAgentID()));
	    
	    // default action dependency: (child=> parent) overriding
	    // (a child by default overrides its parent with a dummy
	    // action named after the child behavior ID)	    
	    addActionDependency(bID, getID().toString(), bID.toString());
	}
    }

    
    /**
     * When this Behavior is no longer a Leaf Behavior, this method adds a connection
     * the Bev(s) one level of hierarchy below this bev. This method is called by
     * the child behavior -- this behavior has no knowledge of the child behavior
     * until the child tells it. 
     *
     * @param bID the <code>BehaviorID</code> of the child behavior
     * @exception RemoteException if an error occurs
     */
    public void addChild(BehaviorID bID) throws RemoteException {
	if (childrenBevs.indexOf(bID) == -1) {
	    childrenBevs.add(bID);
	    // store agent stub
	    agentIDToAgentStubHash.put(bID.getAgentID(), (Behavior)reliesOn(bID.getAgentID()));
	}
    }


    /**
     * When a child behavior of this behavior becomes deactivated, it tells
     * this bev, and this bev disconnects itself from the child. (This method
     * is called by the child behavior.)
     *
     * @param bID a <code>BehaviorID</code> value
     * @exception RemoteException if an error occurs
     */
    public void removeChild(BehaviorID bID) throws RemoteException {
	if (childrenBevs.indexOf(bID) != -1) {
	    childrenBevs.remove(bID);
	}
    }


    /**
     * Returns all parents of this child: i.e., all Behaviors such
     * that a (child=>parent) relationship exists with this behavior
     * (the child). 
     *
     * @return a <code>Vector</code> containing the BehaviorIDs of the
     * parents 
     * @exception RemoteException if an error occurs
     */
    public Vector getParents() throws RemoteException {
	return (Vector) parentBevs.clone();
    }


    /**
     * Returns all parents of this child that are in an active
     * state. Note that if this behavior is active, this method will
     * only return the parents that are in an active state that is
     * hierarchically higher than this behavior. <br>
     *
     * This method is different than <code>getDirectActiveParents()</code>,
     * which returns the parents that are at one level of hierarchy above this
     * bev (in the active behavioral tree).
     *
     * @see #getDirectActiveParents()
     * @return a <code>Vector</code> containing the BehaviorIDs of the
     * parents 
     */
    public Vector getActiveParents() {
	return (Vector) activeParents.clone();
    }


    /**
     * Returns the active parents that are at one level of hierarchy above
     * this bev (in the active behavioral tree). Related to getChildren().
     *
     * @see #getChildren()
     * @return a <code>Vector</code> containing the BehaviorIDs of the parents
     */
    public Vector getDirectActiveParents() {
	return (Vector) activeDirectParents.clone();
    }

    
    /**
     * Returns the behaviors that are at one level of hierarchy below this
     * bev (in the active behavioral tree). Related to getDirectActiveParents().
     *
     * @see #getDirectActiveParents()
     * @return a <code>Vector</code> containing the BehaviorIDs of the children
     */
    public Vector getChildren() {
	return (Vector) childrenBevs.clone();
    }



    /**
     * Adds a parent to its Active Parent collection
     *
     * @param bID the <code>BehaviorID</code> of the parent to be
     * activated 
     */
    public void activateParent(BehaviorID bID) {
	if (activeParents.indexOf(bID) == -1) {
	    activeParents.add(bID);

	    if (isInactive())
		makeListening();
	}
    }
    

    /**
     * Removes a parent from the Active Parent collection
     *
     * @param bID the <code>BehaviorID</code> of the parent to be
     * deactivated 
     */
    public void deactivateParent(BehaviorID bID) {
	if (activeParents.indexOf(bID) != -1) {
	    activeParents.remove(bID);

	    // if this parent was a direct parent,
	    // also remove it from that collection
	    if (isDirectParent(bID)) 
		deactivateDirectParent(bID);

	    if (activeParents.isEmpty())
		makeInactive();
	}
    }

    

    /**
     * Adds a direct parent to its Direct Parent Collection. Is called 
     * internally by the Behavior.
     *
     * @see #getDirectParents() for more information on direct parents
     * @param bID the <code>BehaviorID</code> of the direct parent
     */
    private void activateDirectParent(BehaviorID bID) {
	if (activeDirectParents.indexOf(bID) == -1) {
	    activeDirectParents.add(bID);

	    // tell parent that you're a direct child
	    // (also known as a child... NOT descendent...
	    // we don't keep track of no stinkin descendants)
	    Behavior par = (Behavior) agentIDToAgentStubHash.get(bID.getAgentID());
	    if (par == null) {
		par = (Behavior) reliesOn(bID.getAgentID());
		agentIDToAgentStubHash.put(bID.getAgentID(), par);
	    }

	    try {
		log("DEBUG", "Adding Direct Parent " + bID);
		par.addChild(getID());
	    } catch(Exception e) {
		log("ERROR", "Error trying to connect to parent " + bID);
	    }
	}
    }


    /**
     * Removes a direct parent from its Direct Parent Collection. Is called
     * internally by the Behavior. 
     *
     * @param bID the <code>BehaviorID</code> of the direct parent
     */
    private void deactivateDirectParent(BehaviorID bID) {
	if (activeDirectParents.indexOf(bID) != -1) {
	    activeDirectParents.remove(bID);

	    // tell parent that you -- the direct child --
	    // are no longer active
	    Behavior par = (Behavior) agentIDToAgentStubHash.get(bID.getAgentID());
	    try {
		log("DEBUG", "Removing Direct Parent " + bID);
		par.removeChild(getID());

		// if this was your last direct parent
		// see if you have any more active parents
		// that should now be direct...
		// (i.e., crawl up the tree)
		
		if (!getActiveParents().isEmpty()) {
		    // first make sure that you HAVE active parents

		    while (getDirectActiveParents().isEmpty()) {
			Vector nextLevel = new Vector();
			nextLevel.addAll(par.getDirectActiveParents());
			Enumeration next = nextLevel.elements();
			while (next.hasMoreElements()) {
			    BehaviorID bevID = (BehaviorID) next.nextElement();
			    			    
			    // don't want to look at yourself!
			    if (!bevID.equals(getID())) {
				Behavior bev = (Behavior) agentIDToAgentStubHash.get(bevID.getAgentID());

				if ((isParent(bevID)) && (bev.isActive()))
				// found a direct active parent
				    activateDirectParent(bevID);
				else {
				    nextLevel.addAll(bev.getDirectActiveParents());
				}
			    }
			}

			// next level of bevs ot check out
			next = nextLevel.elements();
		    }
		}		
	    } catch(Exception e) {
		log("ERROR", "Error trying to disconnect from parent " + bID);
	    }
	}
    }
    
    
    /**
     * Whether a behavior is a direct parent of this behavior.
     *
     * @param bID the <code>BehaviorID</code> of the Behavior in question
     * @return true if the behavior is a direct parent
     */
    private boolean isDirectParent(BehaviorID bID) {
	return activeDirectParents.contains(bID);
    }


    /**
     * Whether a behavior is a parent of this behavior.
     *
     * @param bID the <code>BehaviorID</code> of the Behavior in question
     * @return true if the behavior is a parent
     */
    private boolean isParent(BehaviorID bID) {
	return parentBevs.contains(bID);
    }


    /**
     * Adds an action dependency between an action in this behavior
     * and an action in another behavior (parent).
     *
     * @param parID the <code>BehaviorID</code> of the parent behavior
     * @param thisAction the <code>String</code> name of the action in
     * this behavior that is overriding
     * @param parentAction the <code>String</code> name of the action
     * in the parent behavior that is being overriden
     * @exception RemoteException if an error occurs
     */
    // action level dependencies
    public void addActionDependency(BehaviorID parID, String thisAction, String parentAction) throws RemoteException {

	if (!parentToOverridingActionsHash.containsKey(parID)) {
	    // if parent is not registered, register it first
	    // with a hashtable containing this action pair

	    Vector parentActions = new Vector();
	    Hashtable overridingActionsHash = new Hashtable();
	    parentActions.add(parentAction);
	    overridingActionsHash.put(thisAction, parentActions);
	    parentToOverridingActionsHash.put(parID,overridingActionsHash);
	} else {
	    // else, add this action pair to existing hash table
	    Hashtable overs = (Hashtable) parentToOverridingActionsHash.get(parID);

	    // look up any existing dependences with thisAction
	    Vector parActions = (Vector) overs.get(thisAction);
	    if (parActions == null)
		// if there aren't any, create a new vector
		parActions = new Vector();

	    parActions.add(parentAction);
	    overs.put(thisAction, parActions);
	}
    }


    /**
     * Removes an action dependency between an action in this behavior
     * and an action in another behavior (parent).<br>
     *
     * If the parent is not registered with this behavior, returns an
     * error message.
     *
     * @param parID the <code>BehaviorID</code> of the parent behavior
     * @param thisAction the <code>String</code> name of the action in 
     * this behavior that is overriding
     * @param parentAction the <code>String</code> name of this action
     * in the parent behavior that is being overriden
     * @exception RemoteException if an error occurs
     */
    public void deleteActionDependency(BehaviorID parID, String thisAction, String parentAction) throws RemoteException {
	
	if (!parentToOverridingActionsHash.containsKey(parID)) {
	    // if the hashtable doesn't contain the parent key, then 
	    // there isn't anything to delete!!
	    log("ERROR", "Parent BehaviorID was not found!");
	} else {
	    // else, remove this action pair from existing hash table
	    Hashtable overs = (Hashtable) parentToOverridingActionsHash.get(parID);

	    // look up any existing dependencies with thisAction
	    Vector parActions = (Vector) overs.get(thisAction);
	    if (parActions == null)
		// if there aren't any, there's nothing to delete!!
		log("ERROR", "thisAction not found!");
	    else {
		// otherwise, just delete the action pair
		parActions.remove(parentAction);
		overs.put(thisAction, parActions);
	    }
	}
    }




    /**
     * Returns the actions of this Behavior.
     *
     * @return an <code>Enumeration</code> of the actions (as
     * <code>Method</code> objects)
     */
    private Enumeration getActions() {
	return actions.elements();
    }


    /**
     * Returns the names of the actions of this Behavior.
     *
     * @return a <code>Vector</code> containing the names of the
     * action (as <code>String</code> objects)
     */
    public Vector getActionNames() {
	Vector names = new Vector();
	Enumeration actions = getActions();
	while (actions.hasMoreElements()) 
	    names.add(((Method)actions.nextElement()).getName());
	
	// as the list of actions on includes the methods that start
	// with "action", need to artifically add the "activate"
	// action
	names.add("activate");
	
	return names;
    }


    
    /**
     * Returns the names of the actions of this Behavior that have
     * been overriden. (Note that the behavior must be in the overriden
     * state for it to have any overriden actions.
     *
     * @return a <code>Vector</code> of the names of the overriden
     * actions (as <code>String</code> objects)
     */
    public Vector getOverridenActions()  {
	return (Vector) inactiveActions.clone();
    }

    
    /**
     * Returns the names of the actions of this Behavior that are
     * overriden, but for this pass of a secret, are temporarily 
     * unOverriden (i.e., activated) by the behavior that overrode 
     * them in the first place.
     *
     * @see #getOverridenActions()
     * @see #temporarilyActivateAction(String actionName)
     * @return a <code>Vector</code> of the String names of the actions
     */
    private Vector getTemporarilyActiveActions() {
	return (Vector) tempActiveActions.clone();
    }


    /**
     * Returns the names of the actions of this Behavior that can
     * not be overriden by any other behavior. 
     *
     * @see #addPermanentAction(String actionName) 
     * @see #removePermanentAction(String actionName)
     * @return a <code>Vector</code> of the String names of the actions
     */
    public Vector getPermanentActions() {
	return (Vector) permanentActions.clone();
    }


    
    /**
     * Returns the events that this behavior's Actions listens for.
     *
     * @return a <code>Vector</code> of events (notifications) as
     <code>String</code> 
     * @exception RemoteException if an error occurs
     */
    public Vector getEvents() throws RemoteException {
	Vector retVec = new Vector();
	Enumeration evs = eventToActionsVectorHash.keys();
	while (evs.hasMoreElements()) 
	    retVec.add((String)evs.nextElement());
	return retVec;
    }
    

    /**
     * Returns the events that ths behavior's activate method
     * listens for... i.e., the events that will activate this
     * action.
     *
     * @return a <code>Vector</code> of events (notifications) as 
     * <code>String</code>
     * @exception RemoteException if an error occurs
     */
    public Vector getActivationEvents() throws RemoteException {
	return (Vector) activationEvents.clone();
    }


    /**
     * Returns all the events that this behavior listens for,
     * i.e., the combination of getEvents() and getActivationEvents().
     *
     * @return a <code>Vector</code> of events (notifications) as
     * <code>String</code>
     * @exception RemoteException if an error occurs
     */
    public Vector getAllEvents() throws RemoteException {
	Vector allEvents = new Vector();
	allEvents.addAll(getActivationEvents());
	allEvents.addAll(getEvents());
	return allEvents;
    }

    
    /**
     * How a behavior adds its actions. 
     *
     * @param actionName the name of the method to be called (i.e.,
     * the action) as a <code>String</code> 
     */
    
    
    /**
     * How a behavior adds its actions.
     *
     * @param action the method to be called (i.e., the action) as a
     * <code>Method</code> object
     */
    protected void addAction(Method action) {
	if (actions.indexOf(action) == -1)
	    actions.add(action);
    }


    /**
     * Adds an action to the collection of Permanent Actions of this behavior.
     * (A Permanent Action is an action that can not be overriden by any other 
     * behavior.)
     *
     * @param actionName a <code>String</code> value
     */
    public void addPermanentAction(String actionName) {
	if (!permanentActions.contains(actionName))
	    permanentActions.add(actionName);
    }

    
    /**
     * Removes an action from the collection of Permanent Actions of this behavior.
     * (A Permanent Action is an action that can not be overriden by any other 
     * behavior.)
     *
     * @param actionName a <code>String</code> value
     */
    public void removePermanentAction(String actionName) {
	if (permanentActions.contains(actionName))
	    permanentActions.remove(actionName);
    }
    


    /**
     * How a behavior specifies what events (notifications) it wants
     * to listen for, and what actions within the behavior listen for these 
     * events.
     *
     * @param eventName the name of the notification as a <code>String</code> 
     * @param actionName the name of an action that listens for this notification, 
     * as a <code>String</code> 
     */
    protected void addEvent(String eventName, String actionName) {
	if (actionName.equals("activate"))
	    // note that while you can add events for the activate() method,
	    // they are handled differently than events for actions
	    addActivationEvent(eventName);
	else {
	    Object obj = eventToActionsVectorHash.get(eventName);
	    Vector actionsVector;
	    if (obj == null)
		actionsVector = new Vector();
	    else
		actionsVector = (Vector) obj;
	    
	    // convert actionName to the action
	    // then add it to the hashtable
	    actionsVector.add(nameToAction(actionName));
	    eventToActionsVectorHash.put(eventName, actionsVector);
	}
    }    


    /**
     * Adds an activation event. Is only used by the addEvent() method, 
     * for handling activation events separately.
     *
     * @see #addEvent(String eventName, String actionName)
     * @param eventName a <code>String</code> value
     */
    private void addActivationEvent(String eventName) {
	activationEvents.add(eventName);
    }

    
    /**
     * Given the name of an Action in this Behavior, returns the
     * Method object of the Action.
     *
     * @param actionName the name of the Action, as a <code>String</code> object
     * @return the Action, as a <code>Method</code> object
     */
    protected Method nameToAction(String actionName) {
	Enumeration actions = getActions();
	Method returnAction = null;
	while ((actions.hasMoreElements()) && (returnAction == null)) {
	    Method action = (Method) actions.nextElement();
	    String name = action.getName();
	    if (name.equals(actionName)) 
		returnAction = action;
	}
	if (returnAction == null) {
	    log("ERROR", "Method with name " + actionName + " does not exist");
	    return null;
	} else {
	    return returnAction;
	}
    }
	    

    /**
     * The <code>BehaviorAgent</code> itself calls this method to
     * mark an action as overriden. I should probably make this a
     * private / protected method.<br>
     *
     * Note that if the action is permanent, i.e., it can't be overriden,
     * this method logs an error message.
     *
     * @param actionName the name of the action to be overriden
     * @exception RemoteException if an error occurs
     */
    public void overrideAction(String actionName) throws RemoteException {

	// make sure that this action CAN be overriden
	if (!getPermanentActions().contains(actionName)) {
	    inactiveActions.add(actionName);
	    if (!isOverriden())
		makeOverriden();
	} else {
	    log("ERROR", "Action can not be overriden: " + actionName);
	}
    }


    /**
     * The <code>BehaviorAgent</code> itself calls this method to
     * restore an overriden action. I should probably make this a 
     * private / protected method.
     *
     * @param actionName a <code>String</code> value
     * @exception RemoteException if an error occurs
     */
    public void activateAction(String actionName) throws RemoteException {
	inactiveActions.remove(actionName);
	if (inactiveActions.isEmpty()) 
	    makeUnOverriden();
    }


    /**
     * For a given pass of a secret, a behavior can temporarily activate an
     * action (that is overriden) in this action. Note that the bev temporarily
     * activating should be the same bev that overrode the action in the first
     * place. 
     *
     * @see #getTemporarilyActiveActions() 
     * @param actionName the name of the action to temporarily activate, 
     * as a <code>String</code> 
     */
    public void temporarilyActivateAction(String actionName) {
	tempActiveActions.add(actionName);
	log("DEBUG", "...done");
    }

    
    /**
     * Adds the name of a temporarily activated action to this bev's collection.
     * Is called internally by temporarilyActivateAction(String actionName).
     *
     * @see #temporarilyActivateAction(String actionName)
     * @param actionName the name of the action, as a <code>String</code> 
     */
    private void tempActAction(String actionName) {
	//	log("Adding to be temporarily activated: " + actionName);
	inactiveActions.remove(actionName);
    }

    
    /**
     * Removes the name of a temporarily activated action from this bev's collection.
     * Is called internally during the invocation of this bev's actions.<br>
     *
     * If the action to be removed isn't in the temporarily actived actions collection,
     * thie method logs an error. 
     *
     * @see #invokeActiveActions(Vector actions, Secret[] args) 
     * @param actionName the name of the action, as a <code>String</code> value
     */
    private void tempUnActAction(String actionName) {
	//	log("Removing from temporarily activated collection: " + actionName);
	inactiveActions.add(actionName);
	
	if (tempActiveActions.contains(actionName))
	    tempActiveActions.remove(actionName);
	else
	    log("ERROR", "Error Overriding Temporarily Activated Action: " + actionName);
    }


    /**
     * What a child behavior would call to override specific actions
     * of this behavior. The child behavior passes a collection of the
     * names of the actions that it wants to override.
     *
     * @param actionNames a <code>Vector</code> containing the names
     * of the actions that the child behavior wants to override. Each
     * name is a <code>String</code> object.
     * @exception RemoteException if an error occurs
     */
    // called when a child bev is activated
    public void override(Vector actionNames) throws RemoteException {
	// override each action that your child tells you to override

	//	log("I'm being overriden!!");

	Enumeration as = actionNames.elements();
	while (as.hasMoreElements())
	    overrideAction(((String)as.nextElement()));
    }


    /**
     * What a child behavior would call to un-override specific
     * actions of this behavior. The child behavior passes a collection
     * of the names of the actions that it wants to restore.
     *
     * @param actionNames a <code>Vector</code> containing the names
     * of the actions that the child behavior wants to restore. Each
     * name is a <code>String</code> object.
     * @exception RemoteException if an error occurs
     */
    // called when a child bev is deactivated
    public void undoOverride(Vector actionNames) throws RemoteException {
	// undoOverride each action that your child tells you to
	//log("this shit is getting called, yo");
	Enumeration as = actionNames.elements();
	while (as.hasMoreElements())
	    activateAction(((String)as.nextElement()));
    }


    /**
     * Each particular behavior -- i.e., a subclass of this class --
     * needs to implement this method. This method is the Action that is
     * called when the Behavior is Listening; it specifies how the
     * behavior is activated (i.e., what notification(s) activate(s)
     * the behavior).<br>
     * 
     * In this class -- <code>BehaviorAgent</code> -- this method only
     * logs an error message.
     *
     * @param secret the <code>Secret</code> that gets passed to the
     * Behavior. 
     */
    public void activate(Secret secret) {
	log("ERROR","Need to Implement This Method in your Behavior");
    }
    

    /**
     * What the <code>BehaviorCoordinator</code> calls to pass an
     * event (i.e., the Secret) to the Behavior. This method in turn
     * calls all the actions of this Behavior.
     *
     * @see BehaviorCoordinator
     *
     * @param secret the <code>Secret</code> that gets passed
     * @exception RemoteException if an error occurs
     */
    public void pass(Secret secret) throws RemoteException{
	//	log("DEBUG","Bev:...listening.");
	
	//     	log("Secret: " + secret.toString() + "\n");
	//	log("getEvents(): " + getEvents().toString() + "\n");

	// while the number of children who have sent a secret is
	// less than the total number of children, don't handle the secret
	// (this ensures that this bev doesn't handle an event before
	// any of its secrets)
	// (note that if this bev has no children, then it'll just
	// handle the event)
	if (childPasses < getChildren().size()) {
	    log("DEBUG", "This is child number " + childPasses + " to pass this secret to me.");
	    log("DEBUG", "I'll activate on child number " + getChildren().size());
	    childPasses++;
	} else {
	    // all children have handled this event,
	    // so this bev should handle it

	    // reset number of child passes
	    childPasses = 1;	    

	    Secret[] args = new Secret[1];
	    args[0] = secret;
	    
	    if (!isActive()) {
		log("DEBUG", "Waiting to activate");
	    
		//	    log("secret: " + secret.name());
		//	    log("activation events: " + getActivationEvents());


		// check that this behavior can activate on this event
		if (getActivationEvents().indexOf(secret.name()) != -1) {
		    log("DEBUG", "Activating...");
	    
		    // check that it really is listening  
		    // (i.e., that it has ACTIVE parents)
		    
		    Enumeration pars = getActiveParents().elements();
		    
		    while (pars.hasMoreElements()) {
			BehaviorID parID = (BehaviorID) pars.nextElement(); // get bevID
			Behavior par = (Behavior) agentIDToAgentStubHash.get(parID.getAgentID()); // look up agentID to get agent stub
			if (par == null) { 
			    // if agentID wasn't in hashtable, rely on agent and then 
			    // store in hashtable
			    par = (Behavior) reliesOn(parID.getAgentID());
			    agentIDToAgentStubHash.put(parID.getAgentID(), par);
			}
			
			if (!par.isActive()) 
			    deactivateParent(parID);
		    }
		
		    if (!getActiveParents().isEmpty())
			// if it's listening
			// then check activate method
			// (i.e., pass the secret to the activate method)
			activate(secret);
		    else
			makeInactive();
		    
		    // NOTE that you DON't want to pass the secret
		    // to the active parents, since the behavior is
		    // just listening right now, and ISN'T active
		}
	    }
    
	// otherwise, i.e., if bev is active
	// send secret to action methods
	else {
	    // check that this behavior listens for this event
	    if (getEvents().indexOf(secret.name()) != -1) {
		// if bev listens for this event
		// then pass it to its actions
		// and then pass it up to its parents
		
		log("DEBUG", "Handling event");
		 
		// send event to actions registered with that action
		invokeActiveActions((Vector)((Vector)eventToActionsVectorHash.get(secret.name())).clone(), args);
		
		// then send secret to active parents
		Enumeration parents = getActiveParents().elements();
		while (parents.hasMoreElements()) {
		    
		    BehaviorID parID = (BehaviorID) parents.nextElement(); // get bevID
		    Behavior par = (Behavior) agentIDToAgentStubHash.get(parID.getAgentID()); // look up agentID to get agent stub
		    if (par == null) { 
			// if agentID wasn't in hashtable, rely on agent and then 
			// store in hashtable
			par = (Behavior) reliesOn(parID.getAgentID());
			agentIDToAgentStubHash.put(parID.getAgentID(), par);
		    }
		    
		    try {
			// ONLY SEND SECRET TO ACTIVE DIRECT PARENTS
			// (i.e., parents one level of hierarchy above you
			if (isDirectParent(parID))
			    par.pass(secret);
		    } catch(Exception e) {
			log("ERROR", "Error passing secret to parent: " + e.toString());
		    }
		    // if parent is not active (i.e., listening or inactive),
		    // remove it from active parents list
		    if (!par.isActive()) {
			deactivateParent(parID);
			// if this was your last active parent
			// deactivate yourself
			if (getActiveParents().isEmpty()) {
			    makeDeActive();
			    makeInactive();
			} 
		    }
		}

		// if you're currently not active
		// and this is not because you have no more active parents
		// (i.e., you're not active b/c your actionDeactivate was called)
		// then run makeDeActive()
		if (!isActive())
		    makeDeActive();
		

	    } else {
		// if this behavior does not listen for this event
		// then just pass the secret to the parents
		
		log("DEBUG", "Not handling event, just passing it upwards...");
		
		Enumeration parents = getActiveParents().elements();
		while (parents.hasMoreElements()) {
		    BehaviorID parID = (BehaviorID) parents.nextElement(); // get bevID
		    Behavior par = (Behavior) agentIDToAgentStubHash.get(parID.getAgentID()); // look up agentID to get agent stub
		    if (par == null) { 
			// if agentID wasn't in hashtable, rely on agent and then 
			// store in hashtable
			par = (Behavior) reliesOn(parID.getAgentID());
			agentIDToAgentStubHash.put(parID.getAgentID(), par);
		    }
		    
		    try {
			// ONLY SEND SECRET TO ACTIVE DIRECT PARENTS
			// (i.e., parents one level of hierarchy above you
			if (isDirectParent(parID))
			    par.pass(secret);
		    } catch(Exception e) {
			log("ERROR", "Error passing secret to parent: " + e.toString());
		    }

		    // if parent is not active (i.e., listening or inactive),
		    // remove it from active parents list
		    if (!par.isActive()) {
			deactivateParent(parID);
			// if this was your last active parent
			// deactivate yourself
			if (getActiveParents().isEmpty()) {
			    makeDeActive();
			    makeInactive();
			} 
		    }
		}
	    }
	}
	}
    }
	
    

    /**
     * Returns the name of this Behavior. I'll probably get rid of
     * this method soon, as you can do the same thing through the
     * <code>BehaviorID</code>. 
     *
     * @return the name
     * @exception RemoteException if an error occurs
     */
    public String getName() throws RemoteException {
	return getAgentID().getName();
    }


    /**
     * Returns the names of the actions in parent with BehaviorID
     * <code>parID</code> that this behavior overrides.
     *
     * @param parID the <code>BehaviorID</code> of the parent
     * @return the names of all the actions that it overrides (in this
     * parent), as a big <code>String</code>  
     * @exception RemoteException if an error occurs
     */
    public String getDependencies(BehaviorID parID) throws RemoteException {
	Hashtable overs = (Hashtable) parentToOverridingActionsHash.get(parID);
	String actions = new String();
	Vector actionsVector = new Vector();
	if (overs != null) {
	    actions = "\nDependencies for Parent: " + parID + "\n";
	    actions += overs.toString();
	}
	else {
	    actions += "There are no dependencies between this behavior " +
		" and behavior " + parID;
	}
	
	actions += "\n";
	return actions;
    }

    
    /**
     * Returns the names of the actions in all its parents that this
     * behavior overrides.
     *
     * @return the names of all the actions that it overrides, as a
     * big <code>String</code>
     * @exception RemoteException if an error occurs
     */
    public String getDependencies() throws RemoteException {
	Enumeration keys = parentToOverridingActionsHash.keys();
	String result = new String();
	while (keys.hasMoreElements()) 
	    result += getDependencies((BehaviorID) keys.nextElement());
	return result;
    }


    /**
     * How the BehaviorCoordinator registers itself with this
     * Behavior. Is called in the BehaviorCoordinator, after an external
     * Agent has registered this Behavior with the Coordinator.
     *
     * @param aid the <code>AgentID</code> of the BehaviorCoordinator
     * @exception RemoteException if an error occurs
     */
    // so that the bev can access the bevCoord
    public void registerBevCoord(AgentID aid) throws RemoteException {
	bevCoord = (BehaviorCoordinator) reliesOn(aid);
    }

    
    /**
     * Invokes the actions of this bev that are either active or temporarily
     * active. Is called internally by pass(Secret secret).
     *
     * @see #pass(Secret secret) 
     * @param actions a <code>Vector</code> containing the actions of this 
     * bev that are registered for this event
     * @param args the secret (is  a <code>Secret[]</code> object)
     */
    private void invokeActiveActions(Vector actions, Secret[] args) {
	//	log("Invokation Init");
	// invocation initialization
	// (remove temporarily unoverriden actions from overriden actions list)
	Enumeration toaS = getTemporarilyActiveActions().elements();

	//	log("Temporarily activated actions: " + getTemporarilyActiveActions());

	while (toaS.hasMoreElements()) {
	    String toa = (String) toaS.nextElement();
	    tempActAction(toa);
	}

	//	log("Currently overriden actions: " + getOverridenActions());
	// get all the actions
	Enumeration aS = actions.elements();
	while (aS.hasMoreElements()) {
	    Method action = (Method) aS.nextElement();
	    try {
		// for each action, if the action is not currently overriden
		// then invoke it with the current secret
		if (!getOverridenActions().contains(action.getName()))
		    invokeAction(action,args);
	    } catch(Exception e) {
		log("ERROR", "Error invoking active actions: " + e.toString());
	    }
	}

	// invocation termination
	// (add temporarily unoverriden actions back to overriden actions list)
	toaS = getTemporarilyActiveActions().elements();
	while (toaS.hasMoreElements()) {
	    String toa = (String) toaS.nextElement();
	    tempUnActAction(toa);
	}
	//	log("Invokation Termination");
    }


    /**
     * The actual invocation of an action. Is called internally by 
     * invokeActiveActions(Vector actions, Secret[] args). <br>
     *
     * Also keeps track of the current action being invoked, in
     * order to allow unOver() to do its thang. 
     *
     * @see #invokeActiveActions(Vector actions, Secret[] args) 
     * @see #unOver()
     * @param action the action to be invoked, as a <code>Method</code> 
     * @param args the secret, as a <code>Secret[]</code> 
     * @exception InvocationTargetException if an error occurs
     * @exception IllegalAccessException if an error occurs
     */
    private void invokeAction(Method action, Secret[] args) 
	             throws InvocationTargetException, IllegalAccessException {
	currentAction = action.getName(); // keep track of action being invoked
	action.invoke(this, args);
	//	log("Invoking " + action.getName());
	currentAction = new String(); // reset 
    }
    
    
    /**
     * Is called by actions to allow for the temporarily activation of all
     * actions in all ancestor bevs that this action normally overrides. Can
     * be seen as analogous to the <code>super()</code> call in Java, but
     * this analogy doesn't go very far. unOver() doesn't immediately call
     * the action in the ancestor; the temporarily activated action is called
     * when that behavior has the secret. So you can't embed the unOver() call
     * as you would a super() call. Also, unOver() temporarily activates ALL 
     * actions (in ALL bevs) that the current action overrides. (But note that
     * even if this action temporarily activates another action, if other 
     * actions still override the actions, it still won't be fired.) 
     * 
     *
     */
    protected void unOver() {
	// called by actions 
	// (analogous to "super()")


	// the structure of the parentToOverridingActionsHash hashtable
	// (i.e., the method of how the overriding information is stored)
	// calls for the following fairly convoluted proceeding	

	// get all active parents
	Enumeration pars = getActiveParents().elements();
	while (pars.hasMoreElements()) {
	    BehaviorID parID = (BehaviorID) pars.nextElement();

	    // for each parent, get the hashtable that links an action in
	    // in this behavior with actions in the parent that this action
	    // overrides
	    Hashtable overs = (Hashtable) parentToOverridingActionsHash.get(parID);

	    if (overs != null) {
		// get the actions in the parent that 
		// the current action overrides
		Vector actions = (Vector) overs.get(this.currentAction);
		if (actions != null) {
		    Behavior par = (Behavior) agentIDToAgentStubHash.get(parID.getAgentID());
	    
		    Enumeration aS = actions.elements();
		    while (aS.hasMoreElements()) 
			// tell the parent that you want to temporarily
			// activate the overriden actions
			try {
			    String a = (String) aS.nextElement();
			    log("DEBUG", "Want to temporarily activate " + a);
			    par.temporarilyActivateAction(a);
			} catch (Exception e) {
			    log("ERROR", "Error trying to temporarily activate action: " + e.toString());
			}
		}
	    }
	}
    }
	    

    /**
     * Returns a String representation of the agentID-->agentStub mapping.
     *
     * @return the <code>String</code> representation
     */
    public String showHashedAgentStubs() {
	return agentIDToAgentStubHash.toString();
    }

    
    // Speech Stuff
    protected void addGrammar(String gram) {
	try {
	    bevspeech.loadBehaviorGrammar(gram);
	} catch(Exception e) {
	    log("ERROR", "Error trying to add the following behavior grammar: " + gram);
	}
    }

    protected void addKeyNotePair(String key, String note) {
	try {
	    bevspeech.addKeyNotePair(key, note);
	} catch (Exception e) {
	    log("ERROR", "Error adding key note pair: (" + key + "," + note + ")");
	}
    }

    protected void addKeySpokenMsgPair(String key, String spokenMsg) {
	try {
	    bevspeech.addKeySpokenMsgPair(key, spokenMsg);
	} catch (Exception e) {
	    log("ERROR", "Error adding key spokenMsg pair: (" + key + "," + spokenMsg + ")");
	}
    }

    protected void bevSay(String s) {
	try {
	    bevspeech.say(s);
	} catch (Exception e) {
	    log("ERROR", "Error saying \"" + s + "\" : " + e);
	}
    }

    protected void bevSay(String flag, String s) {
	try {
	    bevspeech.say(flag, s);
	} catch (Exception e) {
	    log("ERROR", "Error saying \"" + s + "\" : " + e);
	} 
    }


    // alias stuff
    
    /**
     * Allows designer of a behavior to give his (or her) behavior
     * various aliases.  Adds the string to the collection of this
     * behavior's aliases.  
     *
     * @param String alias
     */
    protected void addAlias(String a) {
	aliases.add(a);
	freeze("aliases", aliases);
    }

    
    /**
     * Returns primary alias of this behavior. 
     *
     * @return the primary alias 
     */
    public String getMainAlias() {
	return (String) aliases.firstElement();
    }

    /**
     * Returns a collection of all aliases of this behavior. 
     *
     * @return the collection of aliases
     */
    public Vector getAliases() {
	return (Vector) aliases.clone();
    }
}








