/**
 * Created by Ajay Kulkarni 6.11.2001
 * 
 * 
 */

package agentland.behavior;

import metaglue.*;
import java.rmi.*;
import java.util.*;
import agentland.util.*;
import agentland.resource.*;
import agentland.behavior.testing.*;


/**
 * The <code>BehaviorCoordinatorAgent</code> is the crux of the Reactive 
 * Behavioral System. It serves as the interface between the outside
 * perceptual world and all the behaviors. A Behavior must be registered
 * with the Behavior Coordinator of a Room to receive information about
 * the Room. <br>
 *
 * The Behavior Coordinator is what keeps track of which bevs are inactive,
 * listening, active leaf, active overriden. It decides which behaviors 
 * receive the events. 
 *
 * @see BehaviorAgent
 * @see ManagedAgent
 * @see BehaviorCoordinator
 */
public class BehaviorCoordinatorAgent extends ManagedAgent implements BehaviorCoordinator {

    private Vector regBevs; // all behaviors registered with the coordinator
    private Vector activeBevs;  // all active behaviors
    private Vector leafBevs; // leaf active behaviors
    private Vector listeningBevs; // listening behaviors
    private Vector inactiveBevs; // inactive behaviors

    // internal counter to keep track of numerics for BehaviorIDs
    private int idCounter;

    private Vector events; // events that the coordinator listens for

    // Look-up Table to translate agentIDs to agent stubs
    private Hashtable agentIDToAgentStubHash = new Hashtable();


    // Methods 


    /**
     * Creates a new <code>BehaviorCoordinatorAgent</code> instance. <br>
     *
     * Adds and activates the Default Behavior Agent. 
     *
     * @exception RemoteException if an error occurs
     */
    public BehaviorCoordinatorAgent() throws RemoteException {
	super();
	
	setLogLevel(LogStream.DEBUG);
	
	regBevs = new Vector();
	activeBevs = new Vector();
	leafBevs = new Vector();
	listeningBevs = new Vector();
	inactiveBevs = new Vector();
	events = new Vector();

	idCounter = 0;
	
	DefaultBehavior defBev = (DefaultBehavior) reliesOn("agentland.behavior.DefaultBehavior");
	addBev(defBev.getID());
	activeBevs.add(defBev.getID());
	leafBevs.add(defBev.getID());
	inactiveBevs.remove(defBev.getID());

	// keep track of agent stub
	// this way, the number of reliesOn() calls are minimized
	agentIDToAgentStubHash.put(defBev.getID().getAgentID(), defBev);
		
	log("DEBUG","BevCoord: Done Booting...");

    }


    /**
     * How a Behavior is registered with the Coordinator. 
     *
     * @param bid the <code>BehaviorID</code> of the Behavior to be
     * registered
     * @exception RemoteException if an error occurs
     */
    public void addBev(BehaviorID bid) throws RemoteException {
	log("DEBUG","Adding Bev " + bid);
	
	Behavior bev = (Behavior)reliesOn(bid.getAgentID());
	Integer num = new Integer(idCounter + 1);
	
	bev.setIDNumeric(num);
	BehaviorID bevID = bev.getID();
	log("DEBUG","Got ID " + bevID);	
	log("DEBUG","Registering Bev");

	if (regBevs.indexOf(bevID) != -1)
	    // behavior is already registered
	    log("Agent is Already Registered");
	else {
	    idCounter++; // only increment counter if you _are_ adding a bev
	    
	    regBevs.add(bevID);
	    inactiveBevs.add(bevID);

	    // check if new bev should be listening
	    // i.e., if at least one of its parents is active	
	    Enumeration parBevIDs = bev.getParents().elements();
	    while (parBevIDs.hasMoreElements()) {
		BehaviorID parID = (BehaviorID) parBevIDs.nextElement();
		if (activeBevs.contains(parID)) {
		    bev.activateParent(parID);
		    if (!listeningBevs.contains(bevID)) {
			listeningBevs.add(bevID);
			inactiveBevs.remove(bevID);
			log("DEBUG","addBev: inactive-->listening");
		    }
		}
	    }    
	    
	    // add thie behavior's events 
	    // (i.e., the notifications that this behavior looks 
	    // out for)
	    // to the behavior coordinator master list
	    Enumeration eS = bev.getAllEvents().elements();
	    while (eS.hasMoreElements()) {
		String eName = (String) eS.nextElement();
		events.add(eName);
		addSpy(eName);
	    }
	    
	    // register the behavior coord with the behavior
	    bev.registerBevCoord(getAgentID());

	    // keep track of agent stubs
	    agentIDToAgentStubHash.put(bevID.getAgentID(), bev);

	    log("Added Bev");
	}

	// finally, after a bev is added to the system, tell the gui
	// (necessary if bevs are added during execution)
	//		notify("behavior.gui.update","");
	// BUT: since we won't dothat in a while, let's just comment
	// that out for now
    }



    /**
     * Returns a String representation of the status of all the bevs 
     * registered with the bev coord. <br>
     *
     * Sample Output: <br>
     * <code>
     * <p>Registered Bevs:<br>
     * ***<br>
     * ***<br>
     * ***<br>
     * ***<br>
     * <p>Inactive Bevs: <br>
     * ***<br>
     * <p>Listening Bevs: <br>
     * ***<br>
     * <p>Active Bevs: <br>
     * ***<br>
     * ***<br>
     * <p>Leaf Active Bevs: <br>
     * ***<br>
     * </code>
     *
     * @return the <code>String</code> representation
     * @exception RemoteException if an error occurs
     */
    public String getBevNames() throws RemoteException {
	String names = new String();

	names += "Registered Behaviors are:\n";
	Enumeration bevs = regBevs.elements();
	if (bevs.hasMoreElements()) { // if there are registered behaviors
	    while (bevs.hasMoreElements()) 
		names += ((BehaviorID) bevs.nextElement()).getName() + "\n";
	    /*
	      names += "\nInactive Behaviors are:\n"; 
	      bevs = inactiveBevs.elements(); 
	      while (bevs.hasMoreElements())  
	      names += ((BehaviorID) bevs.nextElement()).getName() + "\n"; 
	      
	      names += "\nListening Behaviors are:\n"; 
	      bevs = listeningBevs.elements(); 
	      while (bevs.hasMoreElements())  
	      names += ((BehaviorID) bevs.nextElement()).getName() + "\n";
	      
	      names += "\nActive Behaviors are:\n"; 
	      bevs = activeBevs.elements(); 
	      while (bevs.hasMoreElements())  
	      names += ((BehaviorID) bevs.nextElement()).getName() + "\n"; 
	      
	      names += "\nLeaf Active Behaviors are:\n"; 
	      bevs = leafBevs.elements(); 
	      while (bevs.hasMoreElements()) 
	      names += ((BehaviorID) bevs.nextElement()).getName() + "\n"; 
	    */
	} else // if there are no registered behaviors
	    names += "Sorry, there are no registered behaviors.";

	return names;
    }

    public String getActiveBevNames() throws RemoteException {
	String names = new String();

	names += "\nActive Behaviors are:\n";
	Enumeration bevs = activeBevs.elements();
	if (bevs.hasMoreElements()) { // if there are active behaviors
	    while (bevs.hasMoreElements()) 
		names += ((BehaviorID) bevs.nextElement()).getName() + "\n";
	} else 
	    names += "Sorry, there are no active behaviors.";
	
	names += "\nLeaf Active Behaviors are:\n";
	bevs = leafBevs.elements();
	if (bevs.hasMoreElements()) { // if there are leaf behaviors
	    while (bevs.hasMoreElements()) 
		names += ((BehaviorID) bevs.nextElement()).getName() + "\n";
	} else 
	    names += "Sorry, there are no leaf behaviors.";
	
	return names;
    }
    
    public String getInactiveBevNames() throws RemoteException {
	String names = new String();
	
	names += "\nInactive Behaviors are:\n";
	Enumeration bevs = inactiveBevs.elements();
	if (bevs.hasMoreElements()) { // if there are inactive behaviors
	    while (bevs.hasMoreElements()) 
		names += ((BehaviorID) bevs.nextElement()).getName() + "\n";
	} else 
	    names += "Sorry, there are no inactive behaviors.";
	
	return names;
    }

    public String getListeningBevNames() throws RemoteException {
	String names = new String();

	names += "\nListening Behaviors are:\n";
	Enumeration bevs = listeningBevs.elements();
	if (bevs.hasMoreElements()) { // if there are listening behaviors
	    while (bevs.hasMoreElements()) 
		names += ((BehaviorID) bevs.nextElement()).getName() + "\n";
	} else 
	    names += "Sorry, there are no listening behaviors.";
	
	return names;
    }
    
    /**
     * Returns a String Representation of the current active bev tree.
     * <p>Doesn't work yet!
     *
     * @return the <code>String</code> representation
     * @exception RemoteException if an error occurs
     */
    public String getTree() throws RemoteException {
	// er. doesn't quite work. don't use it!!
	
	String header = new String("\nActive Behavioral Tree: \n");
	String tree = new String();
	Vector nextLevel = new Vector();

	Enumeration leafBevs = getLeafBevs().elements();
	
	// while there are more bevs to climb up
	while(true) {
	    // for each level of bevs
	    while (leafBevs.hasMoreElements()) {
		log("TREE: " + tree);

		BehaviorID bevID = (BehaviorID) leafBevs.nextElement();
		Behavior bev = (Behavior) agentIDToAgentStubHash.get(bevID.getAgentID());

		String leaf = bevID.toString();
		nextLevel.addAll(bev.getDirectActiveParents());
		
		tree = leaf + "         " + tree; // 9 spaces
	    }

	    tree = "\n    |||     \n" + tree; 
	    
	    if (nextLevel.isEmpty())
		break;
	    else
		leafBevs = nextLevel.elements();
	}

	return header + tree;
    }
	    
	

    /**
     * Returns a String Representation of the Events that the
     * bev coord listens for. 
     *
     * @return the <code>String</code> representation
     * @exception RemoteException if an error occurs
     */
    public String getEvents() throws RemoteException {
	String names = new String();
	Enumeration eS = events.elements();
	names += "\nEvents:\n";
	while (eS.hasMoreElements()) 
	    names += ((String)eS.nextElement()) + "\n";
	return names;
    }


    
    /**
     * Returns true if the Behavior with BehaviorID bevID is
     * currently active. I'm not sure if it works... I don't really
     * use it. (I should delete it...)
     *
     * @param bevID a <code>BehaviorID</code> value
     * @return a <code>boolean</code> value
     * @exception RemoteException if an error occurs
     */
    public boolean isBevActive(BehaviorID bevID) throws RemoteException {
	if (activeBevs.indexOf(bevID) != -1) 
	    return true;
	else
	    return false;
    }


    /**
     * (This is how the Behavior Coordinator ever gets anything done.) 
     *
     * <p>After getting an event, this method passes it to all the
     * leaf active bevs (which in turn passes it up the active behavioral
     * tree). Then it passes the secret to all the listening bevs, and
     * checks whether any of them get activated. Finally, it cycles through
     * the inactive bevs and checks whether any of them should be listening.
     *
     * @param secret the event to be passed (as a <code>Secret</code>)
     */
    public void tell(Secret secret) {
	log("DEBUG","BevCoord is listening...");

	// need to clone the *Bevs vectors so that 
	// when you remove a bev from it below,
	// it doesn't mess up the Enumeration

	// send secret to leaf active bevs
	Enumeration bevs = ((Vector)leafBevs.clone()).elements();
	try {
	    while (bevs.hasMoreElements()) {
		BehaviorID bevID = (BehaviorID) bevs.nextElement(); // get bevID
		Behavior bev = (Behavior) agentIDToAgentStubHash.get(bevID.getAgentID());
		if (bev == null) {
		    // if agentID wasn't in hashtable, rely on agent and then
		    // store in hashtable
		    bev = (Behavior) reliesOn(bevID.getAgentID());
		    agentIDToAgentStubHash.put(bevID.getAgentID(), bev);
		}

		bev.pass(secret);
	    }
	} catch (RemoteException e) {
	    log("ERROR","Error while passing secret to leaf bevs");
	}

	// send secret to listening bevs
	bevs = ((Vector)listeningBevs.clone()).elements();
	try {
	    while (bevs.hasMoreElements()) {
		BehaviorID bevID = (BehaviorID) bevs.nextElement(); // get bevID
		Behavior bev = (Behavior) agentIDToAgentStubHash.get(bevID.getAgentID());
		if (bev == null) {
		    // if agentID wasn't in hashtable, rely on agent and then
		    // store in hashtable
		    bev = (Behavior) reliesOn(bevID.getAgentID());
		    agentIDToAgentStubHash.put(bevID.getAgentID(), bev);
		}
		
		bev.pass(secret);
	    }
	} catch (RemoteException e) {
	    log("ERROR","Error while passing secret to listening bevs");
	}

	// cycle through inactive bevs to check whether
	// any should be listening
	bevs = ((Vector)inactiveBevs.clone()).elements();
	try {
	    while (bevs.hasMoreElements()) {
		BehaviorID bevID = (BehaviorID) bevs.nextElement(); // get bevID
		Behavior bev = (Behavior) agentIDToAgentStubHash.get(bevID.getAgentID());
		if (bev == null) {
		    // if agentID wasn't in hashtable, rely on agent and then
		    // store in hashtable
		    bev = (Behavior) reliesOn(bevID.getAgentID());
		    agentIDToAgentStubHash.put(bevID.getAgentID(), bev);
		}
		
		Enumeration pars = bev.getParents().elements();
		while (pars.hasMoreElements()) {
		    BehaviorID parID = (BehaviorID) pars.nextElement();
		    if (activeBevs.contains(parID))
			bev.activateParent(parID);
		}
	    }
	} catch (RemoteException e) {
	    log("ERROR","Error while checking inactive bevs");
	}
    }
    
    
    /**
     * A Behavior calls this method to tell the coordinator about
     * a change in its state. 
     *
     * <p>
     * 0 - inactive<br>
     * 1 - listening<br>
     * 2 - active leaf<br>
     * 3 - active, non-leaf (i.e., overriden)<br>
     *
     * @param bevID the <code>BehaviorID</code> of the behavior that's
     * telling the bev coord about the change in its state
     * @param oldstate the old state, as an <code>int</code> value
     * @param newstate the new state, as an <code>int</code> value
     */
    public void updateBev(BehaviorID bevID, int oldstate, int newstate) {
	// first, make sure bev is registered with bev coord
	log("DEBUG", "Want to do (" + oldstate + "--->" + newstate +") for " + bevID); 

	// first, make sure bev is registered with bev coord
	if (regBevs.indexOf(bevID) != -1) {
	    if (oldstate == 0) { // inactive =>
		if (newstate == 1) { // => listening
		    log("DEBUG", "0=>1: " + bevID);
		    listeningBevs.add(bevID);
		    inactiveBevs.remove(bevID);				
		}
		else if (newstate == 2) { // => leaf
		    log("DEBUG", "0=>2: " + bevID);
		    activeBevs.add(bevID);
		    leafBevs.add(bevID);
		    inactiveBevs.remove(bevID);				
		}
	    }
	    else if (oldstate == 1) { // listening =>
		if (newstate == 0) { // => inactive
		    log("DEBUG", "1=>0: " + bevID);
		    inactiveBevs.add(bevID);
		    listeningBevs.remove(bevID);
		}
		else if (newstate == 2) { // => leaf
		    log("DEBUG", "1=>2: " + bevID);
		    activeBevs.add(bevID);
		    leafBevs.add(bevID);
		    listeningBevs.remove(bevID);
		}
	    }
 	    else if (oldstate == 2) { // leaf =>
		if (newstate == 0) { // => inactive
		    log("DEBUG", "2=>0: " + bevID);
		    activeBevs.remove(bevID);
		    leafBevs.remove(bevID);
		    inactiveBevs.add(bevID);
		}
		else if (newstate == 1) { // => listening
		    log("DEBUG", "2=>1: " + bevID);
		    activeBevs.remove(bevID);
		    leafBevs.remove(bevID);
		    listeningBevs.add(bevID);
		}
		else if (newstate == 3) { // => overriden
		    log("DEBUG", "2=>3: " + bevID);
		    leafBevs.remove(bevID);
		}
	    }
	    else if (oldstate == 3) { // overriden =>
		if (newstate == 0) { // => inactive
		    log("DEBUG", "3=>0: " + bevID);
		    activeBevs.remove(bevID);
		    inactiveBevs.add(bevID);
		}
		else if (newstate == 1) { // => listening
		    log("DEBUG", "3=>1: " + bevID);
		    activeBevs.remove(bevID);
		    listeningBevs.add(bevID);
		}
		else if (newstate == 2) { // => leaf
		    log("DEBUG", "3=>2: " + bevID);
		    leafBevs.add(bevID);
		}
	    }
	}    

	// finally, if a bev gets updated,
	// tell the GUI
	notify("behavior.gui.update","");
    }


    /**
     * Returns the collection of registered behaviors. 
     *
     * @return a <code>Vector</code> of BehaviorIDs
     */
    public Vector getRegBevs() {
	return (Vector) regBevs.clone();
    }


    /**
     * Returns the collection of leaf behaviors. 
     *
     * @return a <code>Vector</code> of BehaviorIDs
     */
    public Vector getLeafBevs() {
	return (Vector) leafBevs.clone();
    }


    /**
     * Returns the collection of active behaviors
     *
     * @return a <code>Vector</code> of BehaviorIDs
     */
    public Vector getActiveBevs() {
	return (Vector) activeBevs.clone();
    }


    /**
     * Returns the collection of listening behaviors
     *
     * @return a <code>Vector</code> of BehaviorIDs
     */
    public Vector getListeningBevs() {
	return (Vector) listeningBevs.clone();
    }


    /**
     * Returns the collection of inactive behaviors
     *
     * @return a <code>Vector</code> of BehaviorIDs
     */
    public Vector getInactiveBevs() {
	return (Vector) inactiveBevs.clone();
    }

}



